from typing import Union, Callable, Set, Dict

from gev.utils import import_from_string


def _parse_func(func: Union[str, Callable]) -> Callable:
    return import_from_string(func) if isinstance(func, str) else func


class Event:
    def __init__(self, source: str, type: str, payload: dict):
        self.source = source
        self.type = type
        self.payload = payload

    def __repr__(self) -> str:
        return f'<Event source={self.source!r} type={self.type!r} payload={self.payload!r}>'


class UnknownEventError(Exception):
    pass


class OnSpec:
    def __init__(self, source: str, type: str):
        self.source = source
        self.type = type

    @classmethod
    def from_raw(cls, raw: str):
        """Parse raw string into an OnSpec."""
        source, type = raw.split("::")
        return cls(source=source, type=type)

    def __str__(self):
        return f"{self.source}::{self.type}"

    def __hash__(self) -> int:
        return hash(str(self))


class EventManager:
    _handlers: Dict[str, Set[Callable]]

    def __init__(self):
        self._handlers = {}

    def on(self, spec: str) -> 'ActionDescriptor':
        """
        Register handler for specific type of events.

        Args:
            spec: in format "source::type", e.g. "sys1::step_core_vessel_marked"
        """
        return ActionDescriptor(manager=self, on_spec=spec)

    def take(self, event: Event):
        """Give an event, take actions."""
        on_spec = str(OnSpec(source=event.source, type=event.type))
        if on_spec not in self._handlers:
            raise UnknownEventError(f"No handlers registered for: {on_spec!r}")
        for handler in self._handlers[on_spec]:
            handler(event)

    async def take_async(self, event: Event):
        """Give an event, take actions."""
        on_spec = str(OnSpec(source=event.source, type=event.type))
        if on_spec not in self._handlers:
            raise UnknownEventError(f"No handlers registered for: {on_spec!r}")
        for handler in self._handlers[on_spec]:
            await handler(event)


class ActionDescriptor:
    manager: EventManager
    on_spec: str
    action: Callable

    def __init__(self, manager: EventManager, on_spec: str):
        self.manager = manager
        self.on_spec = on_spec

    def do(self, action: Callable) -> 'EventManager':
        self.manager._handlers.setdefault(self.on_spec, set()).add(
            _parse_func(action)
        )
        return self.manager


default_manager = EventManager()

on = default_manager.on
take = default_manager.take



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gev']

package_data = \
{'': ['*']}

extras_require = \
{'testing': ['pytest>=7.1.2,<8.0.0', 'tox>=3.25.0,<4.0.0']}

setup_kwargs = {
    'name': 'gev',
    'version': '0.1.0',
    'description': 'General Events Manager',
    'long_description': '# Gev (General Events Manager)\n\n```shell\n$ pip install gev\n```\n\n---\n\n## Usage\n\n```python\nfrom gev import Event, EventManager\n\nmanager = EventManager()\n\ndef handler_1(e):\n    print("handler_1 called with", e)\n\ndef handler_2(e):\n    print("handler_2 called with", e)\n\n# Register event handlers\nmanager.on(\'sys_1::event_a\').do(handler_1)\nmanager.on(\'sys_1::event_b\').do(handler_2)\n\nmanager.take(Event(\n    source=\'sys_1\',\n    type=\'event_a\',\n    payload={\'a\': 1}\n))  # handler_1 will be called\n\nmanager.take(Event(\n    source=\'sys_1\',\n    type=\'event_b\',\n    payload={\'b\': 1}\n))  # handler_1 will be called\n```\n\nIf you don\'t want to initialize an `EventManager` instance,\nyou can use the global `default_manager` and its `on` and `take` methods exposed at module level.\n\n```python\nfrom gev import on, take, Event\n\ndef handler_1(e):\n    print("handler_1 called with", e)\n\non(\'sys_1::event_a\').do(handler_1)\n\ntake(Event(\n    source=\'sys_1\',\n    type=\'event_a\',\n    payload={\'a\': 1}\n))  # handler_1 will be called\n```\n',
    'author': 'Wonder',
    'author_email': 'wonderbeyond@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/wonderbeyond/gev',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
